import { Machine, State, On, SideEffect, Goto, Spawn, Unspawn, Interpreter, Interpreter_T, send, Event_T, Context, SideEffectFunction_T } from '../index';

const wait = (ms:number)=>new Promise((resolve)=>{ setTimeout(()=>{ resolve(1); }, ms); });
const makeRequest : SideEffectFunction_T<Cc> = (ctx,e,self)=>{ send(ctx.serverActor, ['received-request',self]); };
const sendResponse : SideEffectFunction_T<Cs> = (ctx,e,self)=>{ send(ctx.clientActor, ['received-response',self]); };
const startTimer : SideEffectFunction_T<Cs> = async (ctx,e,self)=>{ await wait(1500); send(self, ['timer-finished',null]); }
const log : SideEffectFunction_T<Cc|Cs> = (ctx, e, self)=>{ console.log(self.state, ctx); };

type Cc = {
  requestsMade: number;
  responsesReceived: number;
  serverActor: Interpreter_T<Cs>;
};
type Cs = {
  clientActor: Interpreter_T<Cc>
};

const client = 
  Machine<Cc>(
    State<Cc>('idle',
      On<Cc>('entry',
        SideEffect(log),
      ),
      On('server-created',
        SideEffect((_ctx,[_eventName,serverActor],self)=>{ self.context.serverActor=serverActor; }),
        Goto('making-request')
      )
    ),
    State<Cc>('making-request',
      On<Cc>('entry', 
        SideEffect(log),
        SideEffect(makeRequest),
        Context<Cc>((ctx)=>({...ctx, requestsMade: ctx.requestsMade+1})),
        Goto('awaiting-response')
      ),
    ),
    State<Cc>('awaiting-response',
      On<Cc>('entry',
        SideEffect(log),
      ),
      On<Cc>('received-response', 
        SideEffect(log),
        Context<Cc>((ctx)=>({...ctx, responsesReceived: ctx.responsesReceived+1})),
        Goto('making-request')
      ),
    ),
  );

const server = 
  Machine<Cs>(
    State<Cs>('awaiting-request',
      On<Cs>('entry',
        SideEffect(log),
      ),
      On('received-request', 
        SideEffect((_ctx,[_eventName,clientActor],self)=>{ self.context.clientActor=clientActor; }),
        Goto('sending-response')
      ),
    ),
    State<Cs>('sending-response',
      On<Cs>('entry', 
        SideEffect(log),
        SideEffect(startTimer)
      ),
      On('timer-finished',
        SideEffect(sendResponse),
        Goto('awaiting-request')
      )
    ),
  );

const clientActor = Interpreter(client, {context:{requestsMade:0, responsesReceived:0}});
const serverActor = Interpreter(server, {context:{}});
send(clientActor, ['server-created', serverActor]);